# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: echo.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\necho.proto\x12\x04\x64\x65mo\"\x1b\n\x0b\x45\x63hoRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x1c\n\x0c\x45\x63hoResponse\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t2\xd2\x01\n\x04\x45\x63ho\x12/\n\x06\x45\x63hoUU\x12\x11.demo.EchoRequest\x1a\x12.demo.EchoResponse\x12\x31\n\x06\x45\x63hoSU\x12\x11.demo.EchoRequest\x1a\x12.demo.EchoResponse(\x01\x12\x31\n\x06\x45\x63hoUS\x12\x11.demo.EchoRequest\x1a\x12.demo.EchoResponse0\x01\x12\x33\n\x06\x45\x63hoSS\x12\x11.demo.EchoRequest\x1a\x12.demo.EchoResponse(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'echo_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ECHOREQUEST']._serialized_start=20
  _globals['_ECHOREQUEST']._serialized_end=47
  _globals['_ECHORESPONSE']._serialized_start=49
  _globals['_ECHORESPONSE']._serialized_end=77
  _globals['_ECHO']._serialized_start=80
  _globals['_ECHO']._serialized_end=290
# @@protoc_insertion_point(module_scope)
